/* $Id: automat.h,v 1.17 2002/11/18 09:26:30 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifndef AUTOMAT_H
#define AUTOMAT_H

#include <stdlib.h>
#include "lists.h"
/* #include "shared.h" */

typedef struct range
{
	int start;
	int end;
	int repeat;
	int dither;
	int base;
} range;

enum node_types { node_const, node_variable, node_assign,
				  node_binop, node_unop, node_selection, node_rangecheck,
				  node_autoseq, node_autoseqvar /*, node_autoseqstate */
				};

enum node_variable {nv_input, nv_depth, nv_state, nv_topstack,
					nv_arg0 = nv_state, nv_arg1,
					nv_arg2, nv_arg3, nv_arg4, nv_arg5, nv_arg6, 
					nv_arg7, nv_arg8, nv_arg9,
					nv_last = nv_arg9
				   };
				
typedef int reg_block_t[nv_last + 1];
				   
DEFLIST(macrodef, 
		char *name;
		struct node *val;
		);

DEFLIST(rangedef,
		char *name;
		range *r;
		);
				
typedef int (*binop_t)(int, int);

typedef struct node
{
	int type;
	int ref_cnt;
	union
	{
		int val;
		struct
		{
			binop_t opf;
			struct node *left;
			struct node *right;
		} op;
		struct
		{
			range *r;
			struct node *right;
		} sel;
		struct
		{
			int regno;
			struct node *left;
		} a;
		struct
		{
			struct autoseq *seq;
			struct node *right;
			unsigned char reset : 1;
		} aseq;
/*
		struct
		{
			struct autoseq *seq;
			int var;
		} ac;
*/
	} x;
} node;

/* struct autoseqdef; */

int select_value(range *r, int value);
int eval_node(node *n, reg_block_t args);

range *parse_range(char *start, char **end, rangedef **rgs);
int in_range(range *r, int value);
node *parse_expr(char *start, char **end, macrodef *macros, rangedef **rgs);

int eval_expr(char *start, char **end, macrodef *macros, rangedef *rgs);

enum automaton_results { ar_rejected, ar_accepted };

DEFLIST(transition,
		int state;
		node *filter;
		node *stackfilter;
		node *newstate;
		int eat_input;
		node **tostack;
		node **outchars;
		);

#define STATE_PAGE_BITS 7
DEFLIST(automaton, 
	char *name;
	int refcnt;
	macrodef *names;
	rangedef *ranges;
/*	struct autoseqdef *aseqs; */
	int page_cnt;
	transition ***table; // sic! triple!
	transition *ditto;
	int start;
	int fin_cnt;
	int *finals;
);

automaton *create_automaton(char *name);
automaton *load_automaton(char *name);

automaton *get_au_byname(char *name);
void parse_automaton(automaton *au, char *start, char **end);
void destroy_automaton(automaton *au);


enum out_func { aof_normal = 0, aof_state, aof_topstack, aof_depth };

typedef struct auto_state
{
	automaton *au;
	unsigned char accepted : 1;
	int current;
	int depth;
	int maxdepth;
	int *stack;
	// private
	transition *active;
	node **active_out;
	reg_block_t regs;
} auto_state;

typedef void *(*au_ext_create_f)(char *name, char *arg);

typedef int (*au_ext_process_f)(void *id, int *inch, int *outch);

enum au_ext_ops { aeo_reset, aeo_getreg, aeo_isfinal, aeo_mapname, aeo_mapid, aeo_destroy};
typedef int (*au_ext_service_f)(void *id, int op, int ival, char *sval);

DEFLIST(autoseq_branch,
		void *extension;
		auto_state state;
		reg_block_t initial;
		unsigned char master : 1;
		int outfunc;
	);

DEFLIST(autoseq,
		autoseq_branch *branches;
		int inch;
		);

#define AU_PATH_ENV "AU_PATH"
#define AU_START_STACK_DEPTH 64

int is_in_final(auto_state *as);

void autoseq_init(autoseq *seq);
autoseq *parse_autoseq(char *start);
int autoseq_step(autoseq *seq, int *inchar, int *outchar);
int need_feeding(autoseq *seq);
int as_get_reg(autoseq *seq, int reg);
int as_map_name(autoseq *seq, char *name);
char *as_getname_byid(autoseq *seq, int id, char *prefix);
int as_is_in_final(autoseq *seq);
int as_is_accepted(autoseq *seq);
void remove_autoseq(autoseq *as);
autoseq *clone_autoseq(autoseq *src);
void register_au_extension(au_ext_create_f on_create, au_ext_process_f on_process, au_ext_service_f on_service);

#endif // AUTOMAT_H
