/* $Id: uds.c,v 1.6 2002/11/05 10:24:27 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include "util.h"
#include "dstring.h"
#include "lists.h"
#include "interp.h"
#define UDS_IMP
#include "uds.h"

storage_class uds_storage_class = {uds_create, uds_destroy, 
									 uds_check, uds_get,
									 uds_put, uds_add,
									 uds_remove,
									 uds_first, uds_next,
									 uds_enditer,
									 uds_numofitems,
									 uds_seek
									};

typedef struct uds_data
{
	dstring *id;
	interpreter *handler;
	procedure *check_proc, *get_proc, *put_proc, *add_proc,
		*remove_proc, *first_proc, *next_proc, *enditer_proc,
		*numofitems_proc, *seek_proc;
} uds_data;

static int storage_counter;
storage *uds_create(interpreter *i, char *name, char *arg)
{
	storage *ns = ALLOC(storage);
	uds_data *data;
	
	ns->data = data = ALLOC(uds_data);
	data->id = ds_fix(ds_append(ds_createch('S'), ds_fromint(storage_counter++, 10, 0)));
	data->handler = find_subint(i, arg);
	data->check_proc = qfind_proc(data->handler, "check");
	data->get_proc = qfind_proc(data->handler, "get");
	data->put_proc = qfind_proc(data->handler, "put");
	data->add_proc = qfind_proc(data->handler, "insert");
	data->remove_proc = qfind_proc(data->handler, "remove");
	data->first_proc = qfind_proc(data->handler, "first");
	data->next_proc = qfind_proc(data->handler, "next");
	data->enditer_proc = qfind_proc(data->handler, "enditer");
	data->numofitems_proc = qfind_proc(data->handler, "count");
	data->seek_proc = qfind_proc(data->handler, "seek");
	ns->class = &uds_storage_class;
	discard_value(eval_procedurev(data->handler, "create", data->id, NULL));
	return ns;
}

static dstring *get_storage_name(uds_data *uds)
{
	return uds->id;
}

void uds_destroy(interpreter *i, storage *s)
{
	uds_data *ud = s->data;
	discard_value(eval_procedurev(ud->handler, "destroy", get_storage_name(ud), NULL));
}

int uds_check(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	dstring *ret;
	if(!ud->check_proc)
		return 0;
	ret = eval_procedurepv(ud->handler, ud->check_proc, get_storage_name(ud), key, NULL);
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret);
}

dstring *uds_get(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	dstring *ret;
	if(!ud->get_proc)
		return ds_create(NULL);
	ret = eval_procedurepv(ud->handler, ud->get_proc, get_storage_name(ud), key, NULL);
	if(ret)
		ds_unfix(ret);
	return ret ? ret : ds_create(NULL);
}

void uds_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	uds_data *ud = s->data;
	if(!ud->put_proc)
		return;
	discard_value(eval_procedurepv(ud->handler, ud->put_proc, get_storage_name(ud), key, val, NULL));
}

void uds_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	uds_data *ud = s->data;
	if(!ud->add_proc)
		return;
	discard_value(eval_procedurepv(ud->handler, ud->add_proc, get_storage_name(ud), key, val, NULL));
}

void uds_remove(interpreter *i, storage *s, dstring *key)
{
	uds_data *ud = s->data;
	if(!ud->remove_proc)
		return;
	discard_value(eval_procedurepv(ud->handler, ud->remove_proc, get_storage_name(ud), key, NULL));
}

static int unique_uds_enum_id = 1;
dstring *uds_first(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	dstring *ret;
	dstring *enum_id = ds_fromint(unique_uds_enum_id, 10, 0);
	*(int *)iter = unique_uds_enum_id++;
	if(!ud->first_proc)
		return NULL;
	ret = eval_procedurepv(ud->handler, ud->first_proc, get_storage_name(ud), enum_id, NULL);
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

dstring *uds_next(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	dstring *ret;
	if(!ud->next_proc)
		return NULL;
	ret = eval_procedurepv(ud->handler, ud->next_proc, get_storage_name(ud), ds_fromint(*(int *)iter, 10, 0), NULL);
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

void uds_enditer(interpreter *i, storage *s, void **iter)
{
	uds_data *ud = s->data;
	if(ud->enditer_proc)
		discard_value(eval_procedurepv(ud->handler, ud->enditer_proc, get_storage_name(ud), ds_fromint(*(int *)iter, 10, 0), NULL));
}

int uds_numofitems(interpreter *i, storage *s)
{
	uds_data *ud = s->data;
	dstring *ret;
	if(!ud->numofitems_proc)
		return 0;
	ret = eval_procedurepv(ud->handler, ud->numofitems_proc, get_storage_name(ud), NULL);
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? convert_to_int(DS_BODY(ret)) : 0;
}

dstring *uds_seek(interpreter *i, storage *s, void **iter, dstring *key)
{
	uds_data *ud = s->data;
	dstring *ret;
	dstring *enum_id = ds_fromint(unique_uds_enum_id, 10, 0);
	*(int *)iter = unique_uds_enum_id++;
	if(!ud->seek_proc)
		return NULL;
	ret = eval_procedurepv(ud->handler, ud->seek_proc, get_storage_name(ud), enum_id, key, NULL);
	if(ret)
		ds_unfix(ret);
	return ret && !ds_isempty(ret) ? ret : NULL;
}

void uds_init(interpreter *i, storage *s, value *args)
{
	uds_data *ud = s->data;
	value *for_name = ALLOC(value);
	for_name->val = ds_fix(get_storage_name(ud));
	discard_value(eval_procedure(ud->handler, "init", list_add(copy_values(args), for_name)));
}
